#include <common.h>
#include <asm/arch/system.h>
#include <drivers/otp/rtk_otp_debug.h>
#include <drivers/otp/rtk_otp_reg.h>
#include <drivers/otp/rtk_otp_region.h>


#define OTP_BURN_POLLING_TIMES  1000

#define WRITE_REG32(addr, val)        rtd_outl((addr), val)
#define READ_REG32(addr)                rtd_inl((addr))

static spin_lock_t otp_lock = INIT_SPIN_LOCK;


static int rtk_otp_is_readable(unsigned long ofst)
{
    if(ofst <= 0x34 || (ofst <= 0x180 && ofst >= 0x144))
        return -1;
    return 0;
}



int rtk_otp_hw_read(unsigned long ofst, unsigned long *pval)
{
    if (ofst & 0x3) {
        OTP_WARNING("read otp failed, ofst %08lx is not 4 bytes alignments\n", ofst);
        return -1;
    }
    spin_lock(&otp_lock);
    if (rtk_otp_is_readable(ofst) == -1) {
        *pval = 0xdeaddead;
    } else {
        *pval = READ_REG32(OTP_DDR_MAP_BASE + ofst);
    }
    spin_unlock(&otp_lock);
    return 0;

}


int rtk_otp_hw_program(unsigned long ofst, unsigned long val)
{
    int ret = 0;
    unsigned long status;
    int polling_times = OTP_BURN_POLLING_TIMES;
    
    if(ofst & 0x3) {
        OTP_WARNING("write OTP failed, ofst %lx is not 4 bytes alignments\n", ofst);
        return -1;
    }
    if(ofst >= MAX_OTP_OFST) {
        OTP_WARNING("write OTP failed, ofst %lx >= MAX_OTP_OFST(%x)\n", ofst, MAX_OTP_OFST);
        return -1;
    }
    
    if(val == 0)
        return 0;

    spin_lock(&otp_lock);
    
    WRITE_REG32(OTP_PROG_STATUS1, OTP_PROG_STATUS_CLEAN);
    WRITE_REG32(OTP_DDR_MAP_BASE + ofst, val);
    WRITE_REG32(OTP_CTRL_PROG, OTP_CTRL_PROG_EN_PROG_EFUSE);
    do {
        udelay(50);
        status = READ_REG32(OTP_PROG_STATUS1);
        OTP_DBG("OTP_PROG_STATUS1 = %08lx\n", status);
        polling_times--;
        if (polling_times <= 0) {
            if (!(status & OTP_PROG_STATUS_DONE)) {
                OTP_WARNING("Burn OTP Timeout\n");
                ret = -1;
                goto EXIT;
            }
        }
    } while(!(status & OTP_PROG_STATUS_DONE));
    
    if (status & OTP_PROG_STATUS_SUCCESS) {
        OTP_DBG("Program OTP success.\n");
        ret = 0;
    } else {
        if(status & OTP_PROG_STATUS_LOCK) {
            OTP_WARNING("Program OTP Locked.\n");
        } else {
            int idx = 0;
            status = READ_REG32(OTP_PROG_STATUS2);
            OTP_WARNING("ofst = %08lx,Program OTP FAILED, The FAIL bits are:", ofst);
            for (idx = 0; idx < 32; idx++) {
                if(status & (1 << idx))
                    OTP_WARNING("%d ", idx);
            }
            OTP_WARNING(".\n");
            WRITE_REG32(OTP_PROG_STATUS1, OTP_PROG_STATUS_CLEAN);
        } 
        ret = -1;
    }

EXIT:
    spin_unlock(&otp_lock);
    return ret;
}


int rtk_otp_hw_crc(unsigned long start_ofst, unsigned long len, unsigned long *pval)
{
    int ret = -1;
    int times = 200;
    if (start_ofst >= MAX_OTP_OFST || (start_ofst + len) > MAX_OTP_OFST || pval == NULL) {
        OTP_WARNING("params are input error\n");
        return ret;
    }
    spin_lock(&otp_lock);
    
    WRITE_REG32(OTP_CRC, OTP_CRC_CRC_INI);
    OTP_DBG("[%s,%d,%s]0x%08x = 0x%08x\n", __FILE__, __LINE__, __FUNCTION__, OTP_CRC, READ_REG32(OTP_CRC));

    WRITE_REG32(OTP_AUTO_RD_REGION, OTP_AUTO_RD_REGION_START_ADDR(start_ofst)
                | OTP_AUTO_RD_REGION_END_ADDR(start_ofst + len - 1));
    OTP_DBG("[%s,%d,%s]0x%08x = 0x%08x\n", __FILE__, __LINE__, __FUNCTION__, OTP_AUTO_RD_REGION, READ_REG32(OTP_AUTO_RD_REGION));

    WRITE_REG32(OTP_CTRL_CRC_EN, OTP_CRC_EN(1));
    OTP_DBG("[%s,%d,%s]0x%08x = 0x%08x\n", __FILE__, __LINE__, __FUNCTION__, OTP_CTRL_CRC_EN, READ_REG32(OTP_CTRL_CRC_EN));

    WRITE_REG32(OTP_CTRL, (OTP_CTRL_AUTO_RD_REQ | OTP_CTRL_WRITE_EN0));
    OTP_DBG("[%s,%d,%s]0x%08x = 0x%08x\n", __FILE__, __LINE__, __FUNCTION__, OTP_CTRL, READ_REG32(OTP_CTRL));

    while (times --) {
        unsigned long value = READ_REG32(OTP_CTRL);
        if (!(value & OTP_CTRL_AUTO_RD_REQ)) {
            ret = 0;
            break;
        }
        udelay(1000);
    }
    if (ret == 0)
        *pval = GET_OTP_CRC_CRC_OUT(READ_REG32(OTP_CRC));
    WRITE_REG32(OTP_CTRL_CRC_EN, OTP_CRC_EN(0));
    
    spin_unlock(&otp_lock);
    return ret;
}

