/*
 * USB HOST EHCI Controller
 */
#include <common.h>
#include <malloc.h>
#include <usb.h>
#include <watchdog.h>
#include <asm-generic/errno.h>
#include <linux/compat.h>
#include <rbus/usb2_top_reg.h>
#include <rbus/sys_reg_reg.h>
#include "ehci.h"
#include "ehci-plat-config.h"
#include "ehci_reg.h"


#define USB_PHY_DELAY                         120
#define ehci_plat_dbg(fmt, args...)           printk("[ehci], dbg, " fmt, ## args)
#define ehci_plat_info(fmt, args...)          printk("[ehci], info, " fmt, ## args)
#define ehci_plat_warn(fmt, args...)          printk("[ehci], warn, " fmt, ## args)

/*-----------------------------------------------
 * Func : _get_usb2_phy
 *
 * Desc : read usb2 phy register with UTMI
 *
 * Parm : port1     : port index (1~2)
 *        phy_reg   : phy register address
 *
 * Retn : others : phy register value
 *        -1 : failed
 *-----------------------------------------------*/
static int _get_usb2_phy(int port1, int phy_reg)
{
    int phy_value;
    unsigned long addr;

    if(port1 > 2 || port1 < 1) {
        ehci_plat_warn("get_usb2_phy failed, invalid port - %d\n", port1);
        return -1;
    }
    if (phy_reg < 0xe0 || phy_reg > 0xf7) {
        ehci_plat_warn("get_usb2_phy failed, invalid phy_reg 0x%x\n", phy_reg);
        return -1;
    }

    addr = HMAC_HOST_EHCI_INSNREG05_reg;

    rtd_outl(addr, 0x00020000 | (phy_reg & 0x0f) << 8 | ((port1 == 1)? (0x2000) : (0x4000)));
    udelay(USB_PHY_DELAY);

    rtd_outl(addr, 0x00020000 | (phy_reg & 0xd0) << 4 | ((port1 == 1)? (0x2000) : (0x4000)));
    udelay(USB_PHY_DELAY);

    phy_value = rtd_inl(addr) & 0xff;

    return (phy_value & 0xff);
}


/*-----------------------------------------------
 * Func : _set_usb2_phy
 *
 * Desc : set usb2 phy register with UTMI
 *
 * Parm : port1     : port index (1~2)
 *        phy_reg   : phy register address
 *        phy_value : phy register value
 *
 * Retn : 0 : successed,
 *       -1 : failed
 *-----------------------------------------------*/
static int _set_usb2_phy(int port1, int phy_reg, int phy_value) {
    int value;
    unsigned long addr;

    if(port1 > 2 || port1 < 1) {
        ehci_plat_warn("set_usb2_phy failed, invalid port - %d\n", port1);
        return -1;
    }

    if (phy_reg < 0xe0 || phy_reg > 0xf7) {
        ehci_plat_warn("get_usb2_phy failed, invalid phy_reg 0x%x\n", phy_reg);
        return -1;
    }

    addr = HMAC_HOST_EHCI_INSNREG05_reg;

    do {
        // write value
        if (port1 == 1)
            rtd_outl(USB2_TOP_VSTATUS_reg, USB2_TOP_VSTATUS_vstatus_out(phy_value));
        else

        udelay(USB_PHY_DELAY);

        rtd_outl(addr, 0x00020000 | (phy_reg & 0x0f) << 8 | ((port1 == 1)? (0x2000) : (0x4000)));
        udelay(USB_PHY_DELAY);

        rtd_outl(addr, 0x00020000 | (phy_reg & 0xf0) << 4 | ((port1 == 1)? (0x2000) : (0x4000)));
        udelay(USB_PHY_DELAY);

        value = _get_usb2_phy(port1, phy_reg);

        if ((phy_value & 0xff) != value)
            ehci_plat_warn("set_usb2_phy, port %d phy_reg 0x%.2x set 0x%x get 0x%x <---- not matched\n",
                        port1, phy_reg, phy_value & 0xff, value);
        else
            ehci_plat_dbg("set_usb2_phy, port %d phy_reg 0x%.2x set 0x%x get 0x%x\n",
                        port1, phy_reg, phy_value & 0xff, value);

    } while (0); /* just write once */

    return 0;
}

int set_ehci_usb2_phy_reg(unsigned char port1, unsigned char page,
        unsigned char addr, unsigned char val)
{
    int ret = 0;

    if (page > 3) {
        ehci_plat_warn("%s failed, invalid page - %d\n", __func__, page);
        return -ENXIO;
    }
    ret = _set_usb2_phy(port1, 0xF4, (page << 5) | 0x9b); // bit [6:5] : page select = page
    if (ret)
        return ret;

    return _set_usb2_phy(port1, addr, val);
}


int get_ehci_usb2_phy_reg(unsigned char port1, unsigned char page, unsigned char addr)
{
    int ret = 0;

    if (page > 3) {
        ehci_plat_warn("set_u2_phy_reg failed, invalid page - %d\n", page);
        return -ENXIO;
    }

    ret = _set_usb2_phy(port1, 0xF4, (page << 5) | 0x9b);  // bit [6:5] : page select = page
    if (ret)
        return ret;

    return _get_usb2_phy(port1, addr);
}



/*-----------------------------------------------
 * Func : _usb2_load_phy_setting
 *
 * Desc : load usb2 phy setting
 *
 * Parm : p_reg : register list
 *        n_reg : register value
 *
 * Retn : N/A
 *-----------------------------------------------*/
void _usb2_load_phy_setting(
    U2_PHY_REGISTER*    p_reg,
    unsigned int        n_reg
    )
{
    while(n_reg--)
    {
        _set_usb2_phy(p_reg->port1, p_reg->addr, p_reg->val);
        p_reg++;
    }
}

static void ehci_usb2_load_phy_params(void)
{
    ehci_plat_dbg("---- %s start ----\n", __func__);
    _usb2_load_phy_setting(u2_phy_config.p_reg_table, u2_phy_config.n_reg);
    ehci_plat_dbg("---- %s complete ----\n", __func__);
}

/*
 * For MAC_TOP, SNPS reset & release.
 * 1: reset
 * 0: release
*/
static void snps_reset_release(void)
{
    printf("----------- %s ----------- \n", __func__);
    //SNPS reset
    rtd_outl(USB2_TOP_USB_CTR0_reg, USB2_TOP_USB_CTR0_usbphy_pow_p0(1) |
                                    USB2_TOP_USB_CTR0_usbip_port0_soft_reset(1));
    //SNPS release
    rtd_clearbits(USB2_TOP_USB_CTR0_reg, USB2_TOP_USB_CTR0_usbip_port0_soft_reset_mask);
}

static void keep_pll_alive(void) {
    rtd_setbits(USB2_TOP_USB_CTR0_reg, USB2_TOP_USB_CTR0_pll_alive_enable_mask);
}

void ehci_usb_mac_init(struct ehci_hcor *hcor, int index)
{
    int i;
    uint32_t tmp;
    printf("----------- %s ----------- \n", __func__);

    /* UTMI reset */
    if (index == 0) {
        rtd_outl(USB2_TOP_RESET_UTMI_reg, 0x00000000); //MAC 1 PORT 0 UTMI reset
    }

    /* EHCI RUN for Mac 0 and Mac ex */
    rtd_outl((unsigned int)&hcor->or_usbcmd, 0x00080B01);  //bit 0 EHCI Mac EX RUN / set USBCMD
    udelay(400);

    //release owner / Mac-EX//Host software sets this bit as the last action in its process of configuring the Host Controller
    rtd_outl((unsigned int)&hcor->or_configflag, FLAG_CF);  // release owner
    udelay(400);

    /* EHCI port power port ex for mac ex */
    for (i = 0; i < sizeof(hcor->or_portsc)/sizeof(*hcor->or_portsc); i++) {
        tmp = ehci_readl(&hcor->or_portsc[i]);
        ehci_writel(&hcor->or_portsc[i], (tmp | EHCI_PS_PP));
    }
}

static inline void enable_vbus_power(void) {
#ifdef Config_Board_TV006
    // USB_CTL2(MISC_GPIO_BIT_49), USB_CTL3(MISC_GPIO_BIT_48)
    rtd_setbits(0xb8000838, (0xf << 28) | (0xf << 20));  // set pin share to MISC GPIO
    udelay(100);
    rtd_setbits(0xb801bd1c, (0x3 << 16));                // set GPO value
    udelay(100);
    rtd_setbits(0xb801bd04, (0x3 << 16));                // set IO direction to GPO
#endif
}

void usb2_clk_setting_off(void)
{
    printf("----------- %s ----------- \n", __func__);
    /*---------------------------------------------------
    * shutdown sequence :
    *  1. mac clock off
    *  2. not release mac reset
    *  3. not release phy reset
    *---------------------------------------------------*/

    // mac clock disable
    rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_dwc_mask |
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_wrapper_mask |
                SYS_REG_SYS_CLKEN2_write_data(0));


    // mac reset
    rtd_outl(SYS_REG_SYS_SRST2_reg,
                SYS_REG_SYS_SRST2_rstn_usb2_ex_dwc_mask |
                SYS_REG_SYS_SRST2_rstn_usb2_ex_wrapper_mask |
                SYS_REG_SYS_SRST2_write_data(0));


    // phy reset
    rtd_outl(SYS_REG_SYS_SRST3_reg,
                SYS_REG_SYS_SRST3_rstn_usb2_phy_p3_mask |
                SYS_REG_SYS_SRST3_write_data(0));
}


void usb2_clk_setting_on(void)
{
    printf("----------- %s ----------- \n", __func__);
    /*---------------------------------------------------
    * init sequence :
    *  1. mac clock disable
    *  2. mac reset
    *  3. phy reset
    *  4. phy release
    *  5. mac clock enable
    *  6. mac clock disable
    *  7. mac release
    *  8. mac clock enable
    *---------------------------------------------------*/

    /* mac clock disable */
    rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_dwc_mask |
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_wrapper_mask |
                SYS_REG_SYS_CLKEN2_write_data(0));

    /* mac reset */
    rtd_outl(SYS_REG_SYS_SRST2_reg,
                SYS_REG_SYS_SRST2_rstn_usb2_ex_dwc_mask |
                SYS_REG_SYS_SRST2_rstn_usb2_ex_wrapper_mask |
                SYS_REG_SYS_SRST2_write_data(0));


    // phy reset
    rtd_outl(SYS_REG_SYS_SRST3_reg,
                SYS_REG_SYS_SRST3_rstn_usb2_phy_p3_mask |
                SYS_REG_SYS_SRST3_write_data(0));

    // phy reset release
    rtd_outl(SYS_REG_SYS_SRST3_reg,
                SYS_REG_SYS_SRST3_rstn_usb2_phy_p3_mask |
                SYS_REG_SYS_SRST3_write_data(1));

    /* mac clock enable */
    rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_dwc_mask |
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_wrapper_mask |
                SYS_REG_SYS_CLKEN2_write_data(1));


    /* mac clock disable */
    rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_dwc_mask |
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_wrapper_mask |
                SYS_REG_SYS_CLKEN2_write_data(0));


    /* mac reset release */
    rtd_outl(SYS_REG_SYS_SRST2_reg,
                SYS_REG_SYS_SRST2_rstn_usb2_ex_dwc_mask |
                SYS_REG_SYS_SRST2_rstn_usb2_ex_wrapper_mask |
                SYS_REG_SYS_SRST2_write_data(1));


    /* mac clock enable */
    rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_dwc_mask |
                SYS_REG_SYS_CLKEN2_clken_usb2_ex_wrapper_mask |
                SYS_REG_SYS_CLKEN2_write_data(1));

}

unsigned long ehci_base_register[] = {
    HMAC_HOST_ehci_ctrl_start_reg,
};


static int ehci_usb2_set_disconnect_from_otp(void)
{
#define GET_PORT_OFFSET(port)              (28 + (port * 4))
#define GET_PORT_OTP_VAL(val, port)        ((val & (0xf << GET_PORT_OFFSET(port))) >> GET_PORT_OFFSET(port))
#ifndef EFUSE_DATAO_200_reg
#define EFUSE_DATAO_200_reg 0xB8038320
#endif
    int portnum = 0, error = 0;
    unsigned int disconnect_otp = rtd_inl(EFUSE_DATAO_200_reg);
    unsigned int val;

    ehci_plat_info("disconnect_otp=0x%x \n", disconnect_otp);

    //[31:28]=port1
    if (disconnect_otp & 0x1) {
        for (portnum = 0; portnum < U2_PHY_CONFIG_COUNT; portnum++) {
            val = get_ehci_usb2_phy_reg(portnum + 1, 0, 0xe4);
            val &= ~0xf0;
            val |= (GET_PORT_OTP_VAL(disconnect_otp, portnum) << 4);
            error = set_ehci_usb2_phy_reg(portnum + 1, 0, 0xe4, val);
            if (error)
                goto out;
            ehci_plat_info("set port(%d) disconect level to 0x%x \n",
                    (portnum + 1),
                    (get_ehci_usb2_phy_reg(portnum + 1, 0, 0xe4) & 0xf0) >> 4);
        }
    }

out:
    return error;
#undef GET_PORT_OFFSET
#undef GET_PORT_OTP_VAL
}


static int ehci_usb2_set_squelch_from_otp(void)
{
#define GET_PORT_OFFSET(port)              (20 + (port * 4))
#define GET_PORT_OTP_VAL(val, port)        ((val & (0xf << GET_PORT_OFFSET(port))) >> GET_PORT_OFFSET(port))
#ifndef EFUSE_DATAO_182_reg
#define EFUSE_DATAO_182_reg 0xB80382D8
#endif
    int portnum = 0, error = 0;
    unsigned int squelch_otp = rtd_inl(EFUSE_DATAO_182_reg);
    unsigned int val;

    ehci_plat_info("squelch_otp=0x%x\n", squelch_otp);

    //[23:20]=port1, [7]=done
    if (squelch_otp & (0x1<<7)) {
        for (portnum = 0; portnum < U2_PHY_CONFIG_COUNT; portnum++) {
            val = get_ehci_usb2_phy_reg(portnum + 1, 0, 0xe7);
            val &= ~0xf0;
            val |= (GET_PORT_OTP_VAL(squelch_otp, portnum) << 4);
            error = set_ehci_usb2_phy_reg(portnum + 1, 0, 0xe7, val);
            if (error)
                goto out;
            ehci_plat_info("set port(%d) squelch level to 0x%x \n",
                    (portnum + 1),
                    (get_ehci_usb2_phy_reg(portnum + 1, 0, 0xe7) & 0xf0) >> 4);
        }
    }

out:
    return error;

#undef GET_PORT_OFFSET
#undef GET_PORT_OTP_VAL
}


static inline int ehci_usb2_set_phy_from_otp(void)
{
    return (ehci_usb2_set_disconnect_from_otp() ||
            ehci_usb2_set_squelch_from_otp());
}


int ehci_hcd_init(int index, struct ehci_hccr **hccr, struct ehci_hcor **hcor)
{
    printf("----------- %s ----------- \n", __func__);

    if (index >= CONFIG_USB_MAX_CONTROLLER_COUNT) {
        printf("index out of range. index=%d, MAX_HOST_CONTROLLER=%d \n", index, CONFIG_USB_MAX_CONTROLLER_COUNT);
        return -1;
    }

    printf("#@# %s(%d) ehci mac%d\n", __func__, __LINE__, index);

    /* just for mac0 */
    if (index == 0) {
        usb2_clk_setting_on();
        udelay(350);
        snps_reset_release();
        udelay(350);
    }

    keep_pll_alive();

    /* map ehci capacity registers and  operational registers */
    *hccr = (struct ehci_hccr *)ehci_base_register[index];
    *hcor = (struct ehci_hcor *)((uint32_t) *hccr
                + HC_LENGTH(ehci_readl((unsigned int *)&(*hccr)->cr_capbase)));

    debug("ehci: init hccr %x and hcor %x hc_length %d\n",
            (uint32_t)*hccr, (uint32_t)*hcor,
            (uint32_t)HC_LENGTH(ehci_readl((unsigned int *)&(*hccr)->cr_capbase)));

    /* phy setting */
    if (index == 0) {
        ehci_usb2_load_phy_params();
        udelay(100);
        ehci_usb2_set_phy_from_otp();
        udelay(350);
    }

    /* mac setting */
    ehci_usb_mac_init(*hcor, index);
    udelay(350);

    enable_vbus_power();

    return 0;
}

int ehci_hcd_stop(int index)
{
    usb2_clk_setting_off();
    return 0;
}
